/*******************************************************************************
  Controller

  Controller for the MechaRaptor tandem flapping wing system.

  Currently, both ESC are fed the same signal, which means that the two sets of
  wings flap at the same frequency. The phase between the two sets will be
  adjusted manually by setting their initial positions as desired.
  ------------------------------------------------------------------------------
  (c) Copyright 2022 University of Liege
  Author: Thomas Lambert <t.lambert@uliege.be>
  ULiege - Aeroelasticity and Experimental Aerodynamics
  MIT License
  Repo: https://gitlab.uliege.be/thlamb/mecharaptor-controller
*******************************************************************************/


/*******************************************************************************
  Includes
*******************************************************************************/
#include <Servo.h>  // Control ESC by sending appropriate PWM signal

/*******************************************************************************
  Macros (for pins and fixed components only!)
*******************************************************************************/
#define BAUD_RATE 500000  // baud rate for serial (or 115200)

// Pins
#define ESC1_PIN 10  // ESC Front motor
#define ESC2_PIN 9   // ESC Aft motor

#define SWITCH_PIN 7  // ON/OFF switch (mode control)
#define POT_PIN A0    // Potentiometer

#define CLK_PIN 2  // Encoder CLK
#define DO_PIN 8   // Encoder Digital Output
#define CSN_PINS \
  { 6, 5, 4, 3 }  // Encoder Chip Select pins

#define CUR1_PIN A4   // Current measurement for Front motor
#define CUR2_PIN A2   // Current measurement for Aft motor
#define TENS1_PIN A3  // Tension measurement for Front motor
#define TENS2_PIN A1  // Tension measurement for Aft motor

// Values
#define DIV_RES1 48254.0  // Resistance of R1 in the voltage dividers (calibrated so voltage matches)
#define DIV_RES2 6800.0   // Resistance of R2 in the voltage dividers

#define SEN0098_VCC 5                 // Supply voltage of the current sensors
#define SEN0098_SENSI 40.0            // Chip ACS758LCB-050B-PFF-T
#define SEN0098_QUIESCENT_FACTOR 0.5  // Bidirectional chip -> V_IOUTQ = 0.5*VCC
#define SEN0098_OFFSET 0.0122         // Calibrated so current is 0 when no current

#define GEAR_RATIO 0.1  // Gear ratio between motor and wings

// Calibration of encoders
#define ENC_PRECISION 4096  // 12-Bits encoders (nb of points per rot)
#define ENC_ORIENT \
  { 1, -1, 1, -1 }  // Encoder orientation
#define MAX_WING_POS \
  { 222.1, 117.77, 117, 117 }  // Maximum wing angle (returned by encoder value)
#define TRUE_MAX_ANGLE 38.07   // Maximum wing angle (measured on setup)


/*******************************************************************************
  Globals (for things that I can change manually)
 *******************************************************************************/
const boolean gSKIP_CALIB = false;  // Skip calibration or not

// Constant values that can be tweaked here or in the ESC before run
const uint16_t gMAX_SIG = 2000;  // Max signal to the ESC [ms]
const uint16_t gMIN_SIG = 1000;  // Min signal to the ESC [ms]

const uint16_t gMAX_RPM = 3000;    // Motor maximum RPM (set in ESC)
const uint16_t gMIN_RPM = 750;     // Motor minimum RPM (set in ESC)
const uint8_t gMAX_ESC_VAL = 160;  // Value at which motor is at max RPM
const uint8_t gMIN_ESC_VAL = 40;   // Min value for ESC to start motor (trial-error)

const uint8_t gCSN_PINS[] = CSN_PINS;     // Encoder Chip Select pins
const int8_t gENC_ORIENT[] = ENC_ORIENT;  // Encoder orientation

const float gCUR_FACT = SEN0098_SENSI / 1000.0;                 // Factor for current measurement
const float gCUR_QOV = SEN0098_QUIESCENT_FACTOR * SEN0098_VCC;  // VIOUT_Q

// Other
Servo gEsc1;  // ESC for front motor
Servo gEsc2;  // ESC for aft motor

uint8_t gEsc_val = 0;                      // ESC angle (to use with the Servo library) [0-180]
String gMode;                              // Mode of operation
uint16_t gAngleOffset[] = { 0, 0, 0, 0 };  // Encoder offset (prevent wrap-around)
uint16_t gMaxWingPos[] = MAX_WING_POS;     // Encoder maximum recorder position


/*******************************************************************************
  ARDUINO SETUP
 *******************************************************************************/
void setup() {
  Serial.begin(BAUD_RATE);

  // Init ESCs
  Serial.print("[INFO]: Attaching ESCs ... ");
  delay(5000);  // Prevents motor from starting automatically with serial
  gEsc1.attach(ESC1_PIN, gMIN_SIG, gMAX_SIG);
  gEsc2.attach(ESC2_PIN, gMIN_SIG, gMAX_SIG);
  Serial.println("done!");

  // Init switch
  pinMode(SWITCH_PIN, INPUT);

  // Init angle sensors
  for (int i = 0; i < sizeof(gCSN_PINS); i++) {
    pinMode(gCSN_PINS[i], OUTPUT);
    digitalWrite(gCSN_PINS[i], HIGH);
  }
  pinMode(CLK_PIN, OUTPUT);
  pinMode(DO_PIN, INPUT);
  digitalWrite(CLK_PIN, HIGH);

  // Initialize modes
  int switch_state = digitalRead(SWITCH_PIN);

  if (switch_state == LOW) {
    gMode = "MANUAL";
    Serial.println("[WARNING]: Control motors using potentiometer");
  } else {
    gEsc1.write(0);  // Must be initialized at 0 otherwise will not start
    gEsc2.write(0);

    gMode = "SERIAL";
    Serial.println("[WARNING]: Control motors using serial");

    gEsc_val = SetupSerial();
  }

  // Calibrate angle sensors
  if (!gSKIP_CALIB) {
    Calibrate();
  }

  delay(1000);  // Ensure proper initialization
  Serial.println("[INFO]: Setup OK, starting now...");
  delay(500);
}


/*******************************************************************************
  ARDUINO LOOP
  Control the system
  - In serial mode, we impose values based on serial inputs
  - In manual mode, we use the potentiometer
  Acquisition
  - Get values from the 4 rotary encoders
 *******************************************************************************/
void loop() {

  int potent = analogRead(POT_PIN);             // Potentiometer value
  unsigned int wing_angles[sizeof(gCSN_PINS)];  // Wing angles (1L, 1R, 2L, 2R)

  if (gMode == "SERIAL" and Serial.available()) {

    float input = Serial.parseFloat();
    if (input <= 0) {
      StopMotor(gEsc_val);
    } else {
      gEsc_val = FreqToEsc(input);
    }

  } else if (gMode == "MANUAL") {
    gEsc_val = map(potent, 0, 1023, 0, 180);
  }

  // Write value to ESCs
  gEsc1.write(gEsc_val);
  gEsc2.write(gEsc_val);

  // Read value from rotary encoders
  for (int i = 0; i < sizeof(gCSN_PINS); i++) {
    wing_angles[i] = ReadSensor(gCSN_PINS[i], gENC_ORIENT[i], gAngleOffset[i]);
  }

  // Read tension and current from PSU
  float psu_cur[2], psu_vol[2];
  GetPsu(TENS1_PIN, CUR1_PIN, psu_vol[0], psu_cur[0]);
  GetPsu(TENS2_PIN, CUR2_PIN, psu_vol[1], psu_cur[1]);

  // Output all to serial for treatment with serial-studio
  SerialPrint(gMode, potent, gEsc_val, wing_angles, psu_vol, psu_cur);
}


/*******************************************************************************
  OTHER FUNCTIONS
 *******************************************************************************/

// Setup serial
// Get the initial value for the ESCs in serial mode
float SetupSerial() {
  float input;

  float min_freq = RpmToFreq(gMIN_RPM);
  float max_freq = RpmToFreq(gMAX_RPM);

  Serial.print("[INPUT]: Input a flapping frequency between ");
  Serial.print(min_freq);
  Serial.print(" and ");
  Serial.print(max_freq);
  Serial.println(" [Hz]");

  while (input > max_freq or input < min_freq) {
    while (Serial.available() == 0) {}  // Wait for serial input

    input = Serial.parseFloat();
    if (input > max_freq or input < min_freq) {
      Serial.print("[ERROR]: Frequency must be between ");
      Serial.print(min_freq);
      Serial.print(" and ");
      Serial.print(max_freq);
      Serial.print(" [Hz]. Entered: ");
      Serial.println(input);
    }
  }
  Serial.print("[INFO]: Frequency correctly set to ");
  Serial.print(input);
  Serial.println(" Hz.");

  return FreqToEsc(input);
}


// Gracefully stops the motor in serial mode
void StopMotor(int old_esc_val) {

  Serial.println("[INFO] MOTOR STOP INITIATED");

  if (old_esc_val > gMIN_ESC_VAL) {
    Serial.println("[INFO] Slowing down motor for 3 sec...");
    gEsc1.write(gMIN_ESC_VAL);
    gEsc2.write(gMIN_ESC_VAL);
    delay(3000);
  }

  gEsc1.write(0);
  gEsc2.write(0);
  Serial.println("[INFO] Motor stopped completely");

  // Lock system in infinite loop to prevent any restart
  delay(1000);
  while (1) {}
}



// Output data to use with serial-studio
void SerialPrint(String mode, int potent, int esc_val, unsigned int wing_angles[], float psu_vol[], float psu_cur[]) {

  float freq = esc_val * (RpmToFreq(gMAX_RPM) / 180.0);

  Serial.print("/*");  // Beginning of data stream
  Serial.print(mode);
  Serial.print(",");
  Serial.print(millis() / 1000.0);
  Serial.print(",");
  Serial.print(potent * 5.0 / 1023.0);  // Converts pot val into proper tension
  Serial.print(",");
  Serial.print(esc_val);
  Serial.print(",");
  Serial.print(freq);
  for (int i = 0; i < sizeof(gCSN_PINS); i++) {
    Serial.print(",");
    Serial.print(wing_angles[i]);
  }
  for (int i = 0; i < sizeof(gCSN_PINS); i++) {
    Serial.print(",");
    Serial.print(PosToDeg((wing_angles[i] - gMaxWingPos[i]) % ENC_PRECISION) + TRUE_MAX_ANGLE - 360);
  }
  for (int i = 0; i < 2; i++) {
    Serial.print(",");
    Serial.print(psu_vol[i]);
    Serial.print(",");
    Serial.print(psu_cur[i]);
  }
  Serial.println("*/");  // End of data stream
}
